@@ -10,11 +10,12 @@
 //! like Criterion, and for contributors to explore the APIs from
 //! docs.rs.
 
+mod context;
 pub mod dijkstra;
 pub mod files;
 mod graph;
+pub mod hunks;
 pub mod inline;
-mod intervals;
 pub mod line_parser;
 pub mod lines;
 mod positions;
