@@ -2,12 +2,10 @@
 
 #![allow(clippy::mutable_key_type)] // Hash for Syntax doesn't use mutable fields.
 
-use itertools::{EitherOrBoth, Itertools};
 use lazy_static::lazy_static;
 use regex::Regex;
 use std::{
     cell::Cell,
-    cmp::min,
     collections::HashMap,
     env, fmt,
     hash::{Hash, Hasher},
@@ -15,7 +13,7 @@ use std::{
 use typed_arena::Arena;
 
 use crate::{
-    lines::{LineGroup, LineNumber, NewlinePositions},
+    lines::{LineNumber, NewlinePositions},
     positions::SingleLineSpan,
 };
 use ChangeKind::*;
@@ -490,23 +488,42 @@ pub enum TokenKind {
     Atom(AtomKind),
 }
 
+/// A matched token (an atom, a delimiter, or a comment word).
 #[derive(PartialEq, Eq, Debug, Clone)]
 pub enum MatchKind {
     Unchanged {
         highlight: TokenKind,
-        // as this match could be for a list.
+        self_pos: (Vec<SingleLineSpan>, Vec<SingleLineSpan>),
+        // If it's a matched atom, only use the first vec. If it'a
+        // matched delimiter pair, use both vecs.
+        // TODO: better data type.
         opposite_pos: (Vec<SingleLineSpan>, Vec<SingleLineSpan>),
     },
-    Novel,
+    Novel {
+        highlight: TokenKind,
+    },
     UnchangedCommentPart {
+        self_pos: SingleLineSpan,
         opposite_pos: Vec<SingleLineSpan>,
     },
-    ChangedCommentPart,
+    ChangedCommentPart {},
 }
 
 impl MatchKind {
-    pub fn is_unchanged(&self) -> bool {
-        matches!(self, MatchKind::Unchanged { .. })
+    pub fn first_opposite_span(&self) -> Option<SingleLineSpan> {
+        match self {
+            MatchKind::Unchanged { opposite_pos, .. } => opposite_pos.0.first().copied(),
+            MatchKind::UnchangedCommentPart { opposite_pos, .. } => opposite_pos.first().copied(),
+            MatchKind::Novel { .. } => None,
+            MatchKind::ChangedCommentPart {} => None,
+        }
+    }
+
+    pub fn is_change(&self) -> bool {
+        matches!(
+            self,
+            MatchKind::Novel { .. } | MatchKind::ChangedCommentPart {}
+        )
     }
 }
 
@@ -514,14 +531,12 @@ impl MatchKind {
 pub struct MatchedPos {
     pub kind: MatchKind,
     pub pos: SingleLineSpan,
-    // TODO: this is confusing: the previous syntax node with a match
-    // may be on the current line or a previous one.
-    pub prev_opposite_pos: Option<SingleLineSpan>,
 }
 
+// "foo bar" -> vec!["foo", " ", "bar"]
 fn split_words(s: &str) -> Vec<String> {
     lazy_static! {
-        static ref RE: Regex = Regex::new(r"[a-zA-Z0-9]+|[^a-zA-Z0-9]+").unwrap();
+        static ref RE: Regex = Regex::new(r"[a-zA-Z0-9]+|\n|[^a-zA-Z0-9\n]+").unwrap();
     }
 
     RE.find_iter(s).map(|m| m.as_str().to_owned()).collect()
@@ -532,7 +547,6 @@ fn split_comment_words(
     pos: SingleLineSpan,
     opposite_content: &str,
     opposite_pos: SingleLineSpan,
-    prev_opposite_pos: &[SingleLineSpan],
 ) -> Vec<MatchedPos> {
     // TODO: merge adjacent single-line comments unless there are
     // blank lines between them.
@@ -551,13 +565,12 @@ fn split_comment_words(
             diff::Result::Left(word) => {
                 // This word is novel to this side.
                 res.push(MatchedPos {
-                    kind: MatchKind::ChangedCommentPart,
+                    kind: MatchKind::ChangedCommentPart {},
                     pos: content_newlines.from_offsets_relative_to(
                         pos,
                         offset,
                         offset + word.len(),
                     )[0],
-                    prev_opposite_pos: prev_opposite_pos.first().copied(),
                 });
                 offset += word.len();
             }
@@ -573,10 +586,10 @@ fn split_comment_words(
 
                 res.push(MatchedPos {
                     kind: MatchKind::UnchangedCommentPart {
+                        self_pos: word_pos,
                         opposite_pos: opposite_word_pos,
                     },
                     pos: word_pos,
-                    prev_opposite_pos: prev_opposite_pos.first().copied(),
                 });
                 offset += word.len();
                 opposite_offset += opposite_word.len();
@@ -595,8 +608,7 @@ impl MatchedPos {
     fn new(
         ck: ChangeKind,
         highlight: TokenKind,
-        pos: &[SingleLineSpan],
-        prev_opposite_pos: &[SingleLineSpan],
+        pos: (&[SingleLineSpan], &[SingleLineSpan]),
     ) -> Vec<Self> {
         let kind = match ck {
             ReplacedComment(this, opposite) => {
@@ -613,11 +625,10 @@ impl MatchedPos {
 
                 return split_comment_words(
                     this_content,
-                    // TODO: handle the whole pos here.
-                    pos[0],
+                    // TODO: handle the whole pos.0 here.
+                    pos.0[0],
                     opposite_content,
                     opposite_pos[0],
-                    prev_opposite_pos,
                 );
             }
             Unchanged(opposite) => {
@@ -632,27 +643,27 @@ impl MatchedPos {
 
                 MatchKind::Unchanged {
                     highlight,
+                    self_pos: (pos.0.to_vec(), pos.1.to_vec()),
                     opposite_pos,
                 }
             }
-            Novel => MatchKind::Novel,
+            Novel => MatchKind::Novel { highlight },
         };
 
         // Create a MatchedPos for every line that `pos` covers.
+        // TODO: what about opposite pos?
         let mut res = vec![];
-        for (i, line_pos) in pos.iter().enumerate() {
-            // Try to take line N on the opposite side for line N in
-            // this position. If the opposite position is fewer lines,
-            // just take the last line.
-            let opposite_line_pos = prev_opposite_pos
-                .get(i)
-                .or_else(|| prev_opposite_pos.last())
-                .copied();
+        for line_pos in pos.0 {
+            // Don't create a MatchedPos for empty positions. This
+            // occurs when we have lists with empty open/close
+            // delimiter positions, such as the top-level list of syntax items.
+            if line_pos.start_col == line_pos.end_col {
+                continue;
+            }
 
             res.push(Self {
                 kind: kind.clone(),
                 pos: *line_pos,
-                prev_opposite_pos: opposite_line_pos,
             });
         }
 
@@ -670,18 +681,7 @@ pub fn change_positions<'a>(
     let opposite_nl_pos = NewlinePositions::from(opposite_src);
 
     let mut positions = Vec::new();
-    let mut prev_unchanged = vec![SingleLineSpan {
-        line: 0.into(),
-        start_col: 0,
-        end_col: 0,
-    }];
-    change_positions_(
-        &nl_pos,
-        &opposite_nl_pos,
-        nodes,
-        &mut prev_unchanged,
-        &mut positions,
-    );
+    change_positions_(&nl_pos, &opposite_nl_pos, nodes, &mut positions);
     positions
 }
 
@@ -689,7 +689,6 @@ fn change_positions_<'a>(
     nl_pos: &NewlinePositions,
     opposite_nl_pos: &NewlinePositions,
     nodes: &[&Syntax<'a>],
-    prev_opposite_pos: &mut Vec<SingleLineSpan>,
     positions: &mut Vec<MatchedPos>,
 ) {
     for node in nodes {
@@ -706,49 +705,20 @@ fn change_positions_<'a>(
                     .get()
                     .unwrap_or_else(|| panic!("Should have changes set in all nodes: {:#?}", node));
 
-                if let Unchanged(opposite_node) = change {
-                    match opposite_node {
-                        List {
-                            open_position: opposite_open_pos,
-                            ..
-                        } => {
-                            *prev_opposite_pos = opposite_open_pos.clone();
-                        }
-                        Atom { .. } => unreachable!(),
-                    }
-                }
-
                 positions.extend(MatchedPos::new(
                     change,
                     TokenKind::Delimiter,
-                    open_position,
-                    prev_opposite_pos,
+                    (open_position, close_position),
                 ));
 
-                change_positions_(
-                    nl_pos,
-                    opposite_nl_pos,
-                    children,
-                    prev_opposite_pos,
-                    positions,
-                );
+                change_positions_(nl_pos, opposite_nl_pos, children, positions);
 
-                if let Unchanged(opposite_node) = change {
-                    match opposite_node {
-                        List {
-                            close_position: opposite_close_pos,
-                            ..
-                        } => {
-                            *prev_opposite_pos = opposite_close_pos.clone();
-                        }
-                        Atom { .. } => unreachable!(),
-                    }
-                }
                 positions.extend(MatchedPos::new(
                     change,
                     TokenKind::Delimiter,
-                    close_position,
-                    prev_opposite_pos,
+                    // TODO: use open position here too (currently
+                    // breaks display).
+                    (close_position, close_position),
                 ));
             }
             Atom {
@@ -761,39 +731,27 @@ fn change_positions_<'a>(
                     .change
                     .get()
                     .unwrap_or_else(|| panic!("Should have changes set in all nodes: {:#?}", node));
-                if let Unchanged(opposite_node) = change {
-                    match opposite_node {
-                        List { .. } => {
-                            dbg!(node, opposite_node);
-                            unreachable!();
-                        }
-                        Atom {
-                            position: opposite_position,
-                            ..
-                        } => {
-                            *prev_opposite_pos = opposite_position.clone();
-                        }
-                    }
-                }
                 positions.extend(MatchedPos::new(
                     change,
                     TokenKind::Atom(*kind),
-                    position,
-                    prev_opposite_pos,
+                    (position, &[]),
                 ));
             }
         }
     }
 }
 
-fn zip_pad_shorter<Tx: Copy, Ty: Copy>(lhs: &[Tx], rhs: &[Ty]) -> Vec<(Option<Tx>, Option<Ty>)> {
+pub fn zip_pad_shorter<Tx: Clone, Ty: Clone>(
+    lhs: &[Tx],
+    rhs: &[Ty],
+) -> Vec<(Option<Tx>, Option<Ty>)> {
     let mut res = vec![];
 
     let mut i = 0;
     loop {
         match (lhs.get(i), rhs.get(i)) {
             (None, None) => break,
-            (x, y) => res.push((x.copied(), y.copied())),
+            (x, y) => res.push((x.cloned(), y.cloned())),
         }
 
         i += 1;
@@ -802,318 +760,41 @@ fn zip_pad_shorter<Tx: Copy, Ty: Copy>(lhs: &[Tx], rhs: &[Ty]) -> Vec<(Option<Tx
     res
 }
 
-/// Given two slices of line positions, return a list of line number
-/// pairs. If the slices have different lengths, reuse the last item
-/// from the shorter slice.
-fn zip_lines(lhs: &[SingleLineSpan], rhs: &[SingleLineSpan]) -> Vec<(LineNumber, LineNumber)> {
-    let lhs_lines: Vec<_> = lhs.iter().map(|slp| slp.line).collect();
-    let rhs_lines: Vec<_> = rhs.iter().map(|slp| slp.line).collect();
-
-    let lhs_last = match lhs_lines.last() {
-        Some(last) => *last,
-        None => {
-            return vec![];
-        }
+/// Zip `lhs` with `rhs`, but repeat the last item from the shorter
+/// slice.
+pub fn zip_repeat_shorter<Tx: Clone, Ty: Clone>(lhs: &[Tx], rhs: &[Ty]) -> Vec<(Tx, Ty)> {
+    let lhs_last: Tx = match lhs.last() {
+        Some(last) => last.clone(),
+        None => return vec![],
     };
-    let rhs_last = match rhs_lines.last() {
-        Some(last) => *last,
-        None => {
-            return vec![];
-        }
+    let rhs_last: Ty = match rhs.last() {
+        Some(last) => last.clone(),
+        None => return vec![],
     };
 
-    lhs_lines
-        .into_iter()
-        .zip_longest(rhs_lines.into_iter())
-        .map(|l| match l {
-            EitherOrBoth::Both(lhs_line, rhs_line) => (lhs_line, rhs_line),
-            EitherOrBoth::Left(lhs_line) => (lhs_line, rhs_last),
-            EitherOrBoth::Right(rhs_line) => (lhs_last, rhs_line),
-        })
-        .collect()
-}
-
-pub fn aligned_lines(
-    group: &LineGroup,
-    lhs_line_matches: &HashMap<LineNumber, LineNumber>,
-) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let lhs_lines = group.lhs_lines();
-    let rhs_lines = group.rhs_lines();
-
-    // When adding padding to a LineGroup where each side has a
-    // different number of lines, we can end up with extra padding on
-    // the side with fewer lines.
-    //
-    // TODO: fix padding to be smarter.
-    aligned_lines_(&lhs_lines, &rhs_lines, lhs_line_matches)
-}
-
-/// Given two slices of contiguous line numbers, return pairs of
-/// matched lines.
-///
-/// A LHS line is matched with a RHS line if it's present in
-/// `lhs_line_matches` and hasn't already been matched.
-///
-/// If a line has no match on the other side, the pair will contain
-/// None on the other side.
-fn aligned_lines_(
-    lhs_lines: &[LineNumber],
-    rhs_lines: &[LineNumber],
-    lhs_line_matches: &HashMap<LineNumber, LineNumber>,
-) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let mut rhs_highest_matched = rhs_lines.first().map_or(0, |l| l.0 as isize) - 1;
-
-    // For every LHS line, if there is a RHS line that is included in
-    // `rhs_lines` and hasn't yet been paired up, add it to
-    // matched_lines.
-    //
-    // TODO: prefer the line with the most similarity, not just the
-    // first line.  See spurious `let` alignment in 9c71298f8294ce8f,
-    // LHS line 96 in lines.rs.
-    let mut matched_lines = vec![];
-    for lhs_line in lhs_lines {
-        if let Some(rhs_line) = lhs_line_matches.get(lhs_line) {
-            if rhs_line.0 as isize > rhs_highest_matched {
-                matched_lines.push((lhs_line, rhs_line));
-                rhs_highest_matched = rhs_line.0 as isize;
-            }
-        }
-    }
-
     let mut res = vec![];
 
-    let mut lhs_i = 0;
-    let mut rhs_i = 0;
-
-    // Build a vec of matched line tuples. For lines without matches
-    // (novel lines, empty lines), just match lines up pairwise. Pad
-    // gaps if one side has more lines.
-    for (lhs_matched_line, rhs_matched_line) in matched_lines {
-        let mut lhs_prev_lines = vec![];
-        while lhs_i < lhs_lines.len() && lhs_lines[lhs_i] < *lhs_matched_line {
-            lhs_prev_lines.push(lhs_lines[lhs_i]);
-            lhs_i += 1;
-        }
-        let mut rhs_prev_lines = vec![];
-        while rhs_i < rhs_lines.len() && rhs_lines[rhs_i] < *rhs_matched_line {
-            rhs_prev_lines.push(rhs_lines[rhs_i]);
-            rhs_i += 1;
+    let mut i = 0;
+    loop {
+        match (lhs.get(i), rhs.get(i)) {
+            (None, None) => break,
+            (x, y) => res.push((
+                x.cloned().unwrap_or_else(|| lhs_last.clone()),
+                y.cloned().unwrap_or_else(|| rhs_last.clone()),
+            )),
         }
 
-        res.extend(zip_pad_shorter(&lhs_prev_lines, &rhs_prev_lines));
-
-        res.push((Some(*lhs_matched_line), Some(*rhs_matched_line)));
-        lhs_i += 1;
-        rhs_i += 1;
+        i += 1;
     }
 
-    // Handle unmatched lines after the last match.
-    res.extend(zip_pad_shorter(
-        &lhs_lines[min(lhs_i, lhs_lines.len())..],
-        &rhs_lines[min(rhs_i, rhs_lines.len())..],
-    ));
-
     res
 }
 
-pub fn matching_lines<'a>(nodes: &[&Syntax<'a>]) -> HashMap<LineNumber, LineNumber> {
-    let mut res = HashMap::new();
-    for node in nodes {
-        matching_lines_(node, &mut res);
-    }
-    res
-}
-
-fn matching_lines_<'a>(node: &Syntax<'a>, matches: &mut HashMap<LineNumber, LineNumber>) {
-    match node {
-        List {
-            info,
-            open_position,
-            children,
-            close_position,
-            ..
-        } => {
-            if let Some(Unchanged(List {
-                open_position: other_open,
-                close_position: other_close,
-                ..
-            })) = info.change.get()
-            {
-                for (line, other_line) in zip_lines(open_position, other_open) {
-                    matches.entry(line).or_insert(other_line);
-                }
-
-                for (line, other_line) in zip_lines(close_position, other_close) {
-                    matches.entry(line).or_insert(other_line);
-                }
-            }
-
-            for child in children {
-                matching_lines_(child, matches);
-            }
-        }
-        Atom { info, position, .. } => {
-            if let Some(Unchanged(Atom {
-                position: other_pos,
-                ..
-            })) = info.change.get()
-            {
-                for (line, other_line) in zip_lines(position, other_pos) {
-                    matches.entry(line).or_insert(other_line);
-                }
-            }
-        }
-    }
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
     use pretty_assertions::assert_eq;
 
-    #[test]
-    fn test_aligned_middle() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![12.into(), 13.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(2.into(), 12.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![
-                (Some(1.into()), None),
-                (Some(2.into()), Some(12.into())),
-                (None, Some(13.into()))
-            ]
-        );
-    }
-
-    #[test]
-    fn test_aligned_all() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![11.into(), 12.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(1.into(), 2.into());
-        line_matches.insert(2.into(), 12.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![
-                (Some(1.into()), Some(11.into())),
-                (Some(2.into()), Some(12.into())),
-            ]
-        );
-    }
-
-    #[test]
-    fn test_aligned_none() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into()];
-        let rhs_lines: Vec<LineNumber> = vec![11.into()];
-
-        let line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![(Some(1.into()), Some(11.into()))]
-        );
-    }
-
-    #[test]
-    fn test_aligned_line_overlap() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![11.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(1.into(), 11.into());
-        line_matches.insert(2.into(), 11.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![(Some(1.into()), Some(11.into())), (Some(2.into()), None)]
-        );
-    }
-
-    #[test]
-    fn test_aligned_out_of_order() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![11.into(), 12.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(2.into(), 11.into());
-        line_matches.insert(1.into(), 12.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![
-                (None, Some(11.into())),
-                (Some(1.into()), Some(12.into())),
-                (Some(2.into()), None)
-            ]
-        );
-    }
-
-    #[test]
-    fn test_aligned_out_of_range() {
-        let lhs_lines: Vec<LineNumber> = vec![1.into(), 2.into()];
-        let rhs_lines: Vec<LineNumber> = vec![11.into(), 12.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(1.into(), 10.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![
-                (Some(1.into()), Some(11.into())),
-                (Some(2.into()), Some(12.into())),
-            ]
-        );
-    }
-
-    #[test]
-    fn test_aligned_first_line() {
-        let lhs_lines: Vec<LineNumber> = vec![0.into()];
-        let rhs_lines: Vec<LineNumber> = vec![0.into()];
-
-        let mut line_matches: HashMap<LineNumber, LineNumber> = HashMap::new();
-        line_matches.insert(0.into(), 0.into());
-
-        assert_eq!(
-            aligned_lines_(&lhs_lines, &rhs_lines, &line_matches),
-            vec![(Some(0.into()), Some(0.into()))]
-        );
-    }
-
-    /// Ensure that we assign prev_opposite_pos even if the change is on the first node.
-    #[test]
-    fn test_prev_opposite_pos_first_node() {
-        let arena = Arena::new();
-
-        let atom = Syntax::new_atom(
-            &arena,
-            vec![SingleLineSpan {
-                line: 0.into(),
-                start_col: 2,
-                end_col: 3,
-            }],
-            "foo",
-            AtomKind::Normal,
-        );
-        atom.set_change(ChangeKind::Novel);
-        let nodes: Vec<&Syntax> = vec![atom];
-
-        let positions = change_positions("irrelevant", "also irrelevant", &nodes);
-        assert_eq!(
-            positions[0].prev_opposite_pos,
-            Some(SingleLineSpan {
-                line: 0.into(),
-                start_col: 0,
-                end_col: 0
-            })
-        );
-    }
-
     #[test]
     fn test_comment_and_atom_differ() {
         let pos = vec![SingleLineSpan {
@@ -1198,17 +879,16 @@ mod tests {
             end_col: 3,
         };
 
-        let res = split_comment_words(content, pos, opposite_content, opposite_pos, &[]);
+        let res = split_comment_words(content, pos, opposite_content, opposite_pos);
         assert_eq!(
             res,
             vec![MatchedPos {
-                kind: MatchKind::ChangedCommentPart,
+                kind: MatchKind::ChangedCommentPart {},
                 pos: SingleLineSpan {
                     line: 0.into(),
                     start_col: 0,
                     end_col: 3
                 },
-                prev_opposite_pos: None,
             },]
         );
     }
@@ -1219,4 +899,11 @@ mod tests {
         let res = split_words(s);
         assert_eq!(res, vec!["example", ".", "com"])
     }
+
+    #[test]
+    fn test_split_words_treats_newline_separately() {
+        let s = "example.\ncom";
+        let res = split_words(s);
+        assert_eq!(res, vec!["example", ".", "\n", "com"])
+    }
 }
